/*
 * Copyright (c) 2019, 2020, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.graal.python.builtins.objects.object;

import com.oracle.graal.python.PythonLanguage;
import com.oracle.graal.python.builtins.PythonBuiltinClassType;
import com.oracle.graal.python.builtins.objects.PNone;
import static com.oracle.graal.python.builtins.PythonBuiltinClassType.TypeError;
import com.oracle.graal.python.builtins.objects.function.PArguments.ThreadState;
import com.oracle.graal.python.nodes.ErrorMessages;
import com.oracle.graal.python.nodes.PRaiseNode;
import com.oracle.truffle.api.CompilerDirectives;
import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
import com.oracle.truffle.api.dsl.Cached;
import com.oracle.truffle.api.dsl.Cached.Shared;
import com.oracle.truffle.api.dsl.Specialization;
import com.oracle.truffle.api.interop.InteropLibrary;
import com.oracle.truffle.api.interop.UnsupportedMessageException;
import com.oracle.truffle.api.library.CachedLibrary;
import com.oracle.truffle.api.library.ExportLibrary;
import com.oracle.truffle.api.library.ExportMessage;

@ExportLibrary(value = PythonObjectLibrary.class, receiverType = Object.class)
final class DefaultPythonObjectExports {
    @ExportMessage
    static boolean isSequence(Object receiver,
                    @CachedLibrary("receiver") InteropLibrary interopLib) {
        return interopLib.hasArrayElements(receiver);
    }

    @ExportMessage
    static boolean isMapping(Object receiver,
                    @CachedLibrary("receiver") InteropLibrary interopLib) {
        return interopLib.hasMembers(receiver);
    }

    @ExportMessage
    static boolean canBeIndex(Object receiver,
                    @CachedLibrary("receiver") InteropLibrary interopLib) {
        return interopLib.fitsInLong(receiver);
    }

    @ExportMessage
    static Object asIndex(Object receiver,
                    @Shared("raiseNode") @Cached PRaiseNode raise,
                    @CachedLibrary("receiver") InteropLibrary interopLib) {
        if (interopLib.fitsInLong(receiver)) {
            try {
                return interopLib.asLong(receiver);
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        } else if (interopLib.isBoolean(receiver)) {
            try {
                return interopLib.asBoolean(receiver) ? 1 : 0;
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        } else {
            throw raise.raiseIntegerInterpretationError(receiver);
        }
    }

    @ExportMessage
    static int asSize(Object receiver, Object type,
                    @Shared("raiseNode") @Cached PRaiseNode raise,
                    @CachedLibrary(limit = "2") InteropLibrary interopLib) {
        Object index = asIndex(receiver, raise, interopLib);
        if (interopLib.fitsInInt(index)) {
            try {
                return interopLib.asInt(index);
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        } else {
            throw raise.raiseNumberTooLarge(type, index);
        }
    }

    @ExportMessage
    static Object getLazyPythonClass(@SuppressWarnings("unused") Object value) {
        return PythonBuiltinClassType.ForeignObject;
    }

    @ExportMessage
    @TruffleBoundary
    static long hash(Object receiver) {
        return receiver.hashCode();
    }

    @ExportMessage
    static int length(Object receiver,
                    @Shared("raiseNode") @Cached PRaiseNode raise,
                    @CachedLibrary("receiver") InteropLibrary interopLib) {
        if (interopLib.hasArrayElements(receiver)) {
            long sz;
            try {
                sz = interopLib.getArraySize(receiver);
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
            if (sz == (int) sz) {
                return (int) sz;
            } else {
                throw raise.raiseNumberTooLarge(PythonBuiltinClassType.OverflowError, sz);
            }
        } else {
            throw raise.raiseHasNoLength(receiver);
        }
    }

    @ExportMessage
    static class IsTrue {
        @Specialization(guards = "lib.isBoolean(receiver)")
        static boolean bool(Object receiver,
                        @CachedLibrary("receiver") InteropLibrary lib) {
            try {
                return lib.asBoolean(receiver);
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        }

        @Specialization(guards = "lib.fitsInLong(receiver)")
        static boolean integer(Object receiver,
                        @CachedLibrary("receiver") InteropLibrary lib) {
            try {
                return lib.asLong(receiver) != 0;
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        }

        @Specialization(guards = "lib.fitsInDouble(receiver)")
        static boolean floatingPoint(Object receiver,
                        @CachedLibrary("receiver") InteropLibrary lib) {
            try {
                return lib.asDouble(receiver) != 0.0;
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        }

        @Specialization(guards = "lib.hasArrayElements(receiver)")
        static boolean array(Object receiver,
                        @CachedLibrary("receiver") InteropLibrary lib) {
            try {
                return lib.getArraySize(receiver) > 0;
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        }

        @Specialization(guards = {
                        "!lib.isBoolean(receiver)", "!lib.fitsInLong(receiver)",
                        "!lib.fitsInDouble(receiver)", "!lib.hasArrayElements(receiver)"
        })
        static boolean generic(Object receiver,
                        @CachedLibrary("receiver") InteropLibrary lib) {
            return !lib.isNull(receiver);
        }
    }

    @ExportMessage
    static boolean isSame(Object receiver, Object other,
                    @CachedLibrary("receiver") InteropLibrary receiverLib,
                    @CachedLibrary(limit = "3") InteropLibrary otherLib) {
        return receiverLib.isIdentical(receiver, other, otherLib);
    }

    @ExportMessage
    @SuppressWarnings("unused")
    static int equalsInternal(Object receiver, Object other, ThreadState threadState,
                    @CachedLibrary("receiver") InteropLibrary receiverLib,
                    @CachedLibrary(limit = "3") InteropLibrary otherLib) {
        return receiverLib.isIdentical(receiver, other, otherLib) ? 1 : 0;
    }

    @ExportMessage
    static boolean equalsWithState(Object receiver, Object other, PythonObjectLibrary oLib, ThreadState state,
                    @CachedLibrary("receiver") InteropLibrary receiverLib,
                    @CachedLibrary(limit = "3") InteropLibrary otherLib) {
        return receiverLib.isIdentical(receiver, other, otherLib) || oLib.equalsInternal(receiver, other, state) == 1;
    }

    @ExportMessage
    static boolean isForeignObject(Object receiver,
                    @CachedLibrary("receiver") InteropLibrary lib) {
        try {
            return !lib.hasLanguage(receiver) || lib.getLanguage(receiver) != PythonLanguage.class;
        } catch (UnsupportedMessageException e) {
            // cannot happen due to check
            CompilerDirectives.transferToInterpreterAndInvalidate();
            throw new IllegalStateException(e);
        }
    }

    @ExportMessage
    static Object asPString(Object receiver,
                    @CachedLibrary(limit = "1") InteropLibrary lib,
                    @Cached.Exclusive @Cached PRaiseNode raise) {
        if (lib.isString(receiver)) {
            try {
                return lib.asString(receiver);
            } catch (UnsupportedMessageException e) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(e);
            }
        }
        throw raise.raise(PythonBuiltinClassType.TypeError, ErrorMessages.EXPECTED_STR_BYTE_OSPATHLIKE_OBJ, receiver);
    }

    @ExportMessage
    static boolean canBeJavaDouble(@SuppressWarnings("unused") Object receiver,
                    @CachedLibrary(limit = "1") InteropLibrary interopLib) {
        return interopLib.fitsInDouble(receiver);
    }

    @ExportMessage
    static double asJavaDouble(Object receiver,
                    @Cached.Exclusive @Cached PRaiseNode raise,
                    @CachedLibrary(limit = "1") InteropLibrary interopLib) {
        if (canBeJavaDouble(receiver, interopLib)) {
            try {
                return interopLib.asDouble(receiver);
            } catch (UnsupportedMessageException ex) {
                // cannot happen due to check
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(ex);
            }
        }
        throw raise.raise(TypeError, ErrorMessages.MUST_BE_REAL_NUMBER, receiver);

    }

    @ExportMessage
    public static Object lookupAttribute(@SuppressWarnings("unused") Object x, @SuppressWarnings("unused") String name, @SuppressWarnings("unused") boolean inheritedOnly) {
        return PNone.NO_VALUE;
    }

    @ExportMessage
    static boolean canBeJavaLong(@SuppressWarnings("unused") Object receiver,
                    @CachedLibrary(limit = "1") InteropLibrary interopLib) {
        return interopLib.fitsInLong(receiver);
    }

    @ExportMessage
    static long asJavaLong(Object receiver,
                    @Cached.Exclusive @Cached PRaiseNode raise,
                    @CachedLibrary(limit = "1") InteropLibrary interopLib) {
        if (canBeJavaDouble(receiver, interopLib)) {
            try {
                return interopLib.asLong(receiver);
            } catch (UnsupportedMessageException ex) {
                // cannot happen due to check
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(ex);
            }
        }
        throw raise.raise(TypeError, ErrorMessages.MUST_BE_REAL_NUMBER, receiver);
    }

    @ExportMessage
    static boolean canBePInt(@SuppressWarnings("unused") Object receiver,
                    @CachedLibrary("receiver") InteropLibrary lib) {
        return lib.fitsInLong(receiver);
    }

    @ExportMessage
    static long asPInt(Object receiver,
                    @CachedLibrary("receiver") InteropLibrary lib,
                    @Cached.Exclusive @Cached PRaiseNode raise) {
        if (lib.fitsInLong(receiver)) {
            try {
                return lib.asLong(receiver);
            } catch (UnsupportedMessageException ex) {
                // cannot happen due to check
                CompilerDirectives.transferToInterpreterAndInvalidate();
                throw new IllegalStateException(ex);
            }
        }
        throw raise.raise(TypeError, ErrorMessages.OBJ_CANNOT_BE_INTERPRETED_AS_INTEGER, receiver);
    }

}
